#!/usr/bin/python
# -*- coding: iso-8859-15 -*-

#
#                       PyMP.mdct.block
#
#
#
# M. Moussallam                             Created on Nov 12, 2012
# -----------------------------------------------------------------------


"""

Module mdct.block
=================

This class inherits from :class:`.BaseBlock` and is used to represent and manipulate MDCT atoms.
:class:`mdct.Atom` Objects can either be constructed or recovered from Xml using :func:`mdct.Atom.fromXml`


MDCT Blocks are the objects encapsulating the actual atomic projections and the selection step of MP

This module describes 3 kind of blocks:
    - :class:`Block`  is a block based on the standard MDCT transform. Which means atoms localizations
                 are constrained by the scale of the transform
                 The atom selected is simple the one that maximizes the correlation with the residual
                 it is directly indexes by the max absolute value of the MDCT bin
                 No further optimization of the selected atom is performed

    - :class:`LOBlock`   is a block that performs a local optimization of the time localization of the selected atom

    - :class:`FullBlock`  this block simulates a dictionary where atoms at all time localizations are available
                  This kind of dictionary can be thought of as a Toeplitz matrix
                  BEWARE: memory consumption and CPU load will be very high! only use with very small signals
                  (e.g. 1024 samples or so) Fairly intractable at higher dimensions
"""

import sys

#from numpy.fft import fft
import numpy as np

from math import  floor
from cmath import exp
#import matplotlib.pyplot as plt

from .. import win_server
from .. import  log
#from ..tools import Xcorr  # , Misc
from .. import parallelProjections
from ..base import BaseBlock
from . import atom



# declare global PyWinServer shared by all MDCT blocks instances
#global _PyServer
#, _Logger
_PyServer = win_server.get_server()
_Logger = log.Log('MDCTBlock', level=0)


class Block(BaseBlock):
    """ This class is a basic MDCT block. It handles a set of routines that allows greedy
        decomposition algorithms (MP, OMP , GP ...) to iteratively project the ongoing residual
        onto a specified MDCT base and to retrieve the maximum inner product

         IMPORTANT: the computation of the inner products relies on a C-written module parallelProjections
                 that needs being compiled first on whatever your plateform.
                 Run the test file: parallelProjections/install_and_test.py in order to compile it
                 and perform some basic execution tests

         """

    #members
    nature = 'MDCT'
    framed_data_matrix = None
    projs_matrix = None
    frame_len = 0
    frame_num = 0

    max_index = 0
    max_value = 0
    max_bin_idx = 0
    max_frame_idx = 0

    # MDCT static window and twiddle parameters
    w_long = None
    w_ledge = None
    w_redge = None

    pre_twid_vec = None
    post_twid_vec = None

    # store fft matrix for later Xcorr purposes
    fftMat = None
    fft = None
    # Score tree
    best_score_tree = None

    # optim?
    use_c_optim = True
    
    # DEPRECATED
    HF = False
    HFlimit = 0.1

    windowType = None
    
    # @TODO REMOVE: ONLY FOR BENCHMARKING
    nb_up_frame = 0
    # constructor - initialize residual signal and projection matrix
    def __init__(self, length=0, resSignal=None, frameLen=0, useC=True, forceHF=False, debug_level=None):
        if debug_level is not None:
            _Logger.set_level(debug_level)

        self.scale = length
        self.residual_signal = resSignal

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len
        self.projs_matrix = np.zeros(len(self.framed_data_matrix))
        self.use_c_optim = useC
        self.HF = forceHF
        _Logger.info('new MDCT block constructed size : ' + str(self.scale))
#        self.projs_matrix = zeros((self.frameNumber , self.frameLength))

    # compute mdct of the residual and instantiate various windows and twiddle
    # coefficients
    def initialize(self):
        """ Compute mdct of the residual and instantiate various windows and twiddle coefficients"""
        #Windowing
        L = self.scale

        self.w_long = np.array([np.sin(float(l + 0.5) * (np.pi / L)) for l in range(L)])

        # twidlle coefficients
        self.pre_twid_vec = np.array([exp(n * (-1j) * np.pi / L) for n in range(L)])
        self.post_twid_vec = np.array(
            [exp((float(n) + 0.5) * -1j * np.pi * (L / 2 + 1) / L) for n in range(L / 2)])

        if self.windowType == 'half1':
            self.w_long[0:L / 2] = 0
            # twidlle coefficients
            self.pre_twid_vec[0:L / 2] = 0
#        self.fftMat = zeros((self.scale , self.frameNumber) , complex)
#        self.normaCoeffs = sqrt(1/float(L))

        # score tree - first version simplified
        self.best_score_tree = np.zeros(self.frame_num)

        if self.HF:
            self.bestScoreHFTree = np.zeros(self.frame_num)

        # OPTIM -> do pre-twid directly in the windows
        self.locCoeff = self.w_long * self.pre_twid_vec

    # Search among the inner products the one that maximizes correlation
    # the best candidate for each frame is already stored in the best_score_tree
    def find_max(self):
        """Search among the inner products the one that maximizes correlation
        the best candidate for each frame is already stored in the best_score_tree """
        treeMaxIdx = self.best_score_tree.argmax()

        maxIdx = abs(self.projs_matrix[treeMaxIdx * self.scale /
            2: (treeMaxIdx + 1) * self.scale / 2]).argmax()
        self.maxIdx = maxIdx + treeMaxIdx * self.scale / 2
        self.max_value = self.projs_matrix[self.maxIdx]

#        print "block find_max called : " , self.maxIdx , self.max_value

        if self.HF:
            treemaxHFidx = self.bestScoreHFTree.argmax()
            maxHFidx = abs(self.projs_matrix[(treemaxHFidx + self.HFlimit)
                * self.scale / 2: (treemaxHFidx + 1) * self.scale / 2]).argmax()

            self.maxHFIdx = maxHFidx + (
                treemaxHFidx + self.HFlimit) * self.scale / 2
            self.maxHFValue = self.projs_matrix[self.maxHFIdx]

    # construct the atom that best correlates with the signal
    def get_max_atom(self, HF=False):
        """ construct the atom that best correlates with the signal"""

        if not HF:
            self.max_frame_idx = floor(self.maxIdx / (0.5 * self.scale))
            self.max_bin_idx = self.maxIdx - self.max_frame_idx * (
                0.5 * self.scale)
        else:
            self.max_frame_idx = floor(self.maxHFIdx / (0.5 * self.scale))
            self.max_bin_idx = self.maxHFIdx - self.max_frame_idx * (
                0.5 * self.scale)
        Atom = atom.Atom(self.scale, 1, max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0), self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx
#        print self.max_bin_idx, Atom.reducedFrequency
        if not HF:
            Atom.mdct_value = self.max_value
        else:
            Atom.mdct_value = self.maxHFValue

        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom()

        if HF:
            Atom.waveform = Atom.waveform * (self.maxHFValue / self.max_value)
        return Atom

    #
    def update(self, new_res_signal, startFrameIdx=0, stopFrameIdx=-1):
        """ update the part of the residual that has been changed and update inner products    """
# print "block update called : " , self.scale , new_res_signal.length ,
# self.frameNumber
        self.residual_signal = new_res_signal

        if stopFrameIdx < 0:
            endFrameIdx = self.frame_num - 1
        else:
            endFrameIdx = stopFrameIdx

        L = self.scale

        self.framed_data_matrix[startFrameIdx * L / 2: endFrameIdx * L / 2 + L] = self.residual_signal.data[startFrameIdx * self.frame_len: endFrameIdx * self.frame_len + 2 * self.frame_len]

        self.compute_transform(startFrameIdx, stopFrameIdx)

        self.find_max()

    # inner product computation through MDCT
    def compute_transform(self, startingFrame=1, endFrame=-1):
        """ inner product computation through MDCT """
        if self.w_long is None:
            self.initialize()

        if endFrame < 0:
            endFrame = self.frame_num - 1

        # debug -> changed from 1: be sure signal is properly zero -padded
        if startingFrame < 1:
            startingFrame = 1

        # @TODO REMOVE: ONLY FOR BENCHMARKING
        self.nb_up_frame += endFrame-startingFrame
        # Wrapping C code call for fast implementation
#        if self.use_c_optim:
        try:
            parallelProjections.project(self.framed_data_matrix, self.best_score_tree,
                                             self.projs_matrix,
                                             self.locCoeff,
                                             self.post_twid_vec,
                                             startingFrame,
                                             endFrame,
                                             self.scale, 0)

        except SystemError:
            print sys.exc_info()[0]
            print sys.exc_info()[1]
            raise
        except:
            print "Unexpected error:", sys.exc_info()[0]
            raise
#        else:
#            try:
#                import fftw3
#            except ImportError:
#                print '''Impossible to load fftw3, you need to use the C extension library or have
#                        a local installation of the python fftw3 wrapper '''
#                return
#        # create a forward fft plan, since fftw is not faster for computing
#        # DCT's no need to use it for that
#            L = self.scale
#            K = L / 2
#            T = K / 2
#            normaCoeffs = sqrt(2 / float(K))
#            if self.fft is None:
#                self.inputa = self.framed_data_matrix[K -
#                     T: K + L - T].astype(complex)
#                self.outputa = None
#                self.fft = fftw3.Plan(self.inputa, self.
#                    outputa, direction='forward', flags=['estimate'])
#
#            self.project(startingFrame, endFrame, L, K, T, normaCoeffs)

        if self.HF:
            for i in range(startingFrame, endFrame):
                self.bestScoreHFTree[i] = abs(self.projs_matrix[
                    (i + self.HFlimit) * self.scale / 2: (i + 1) * self.scale / 2]).max()

    # UPDATE: need to keep this slow version for non C99 compliant systems
    # (windows)
#    def project(self, startingFrame, endFrame, L,  K, T, normaCoeffs):
#
#        for i in range(startingFrame, endFrame):
#            # initialize input data
#            self.inputa[:] = 0
#            self.inputa += self.framed_data_matrix[i * K - T: i *
#                K + L - T] * self.locCoeff
#
#            #compute fft
#            self.fft()
#
#            # post-twiddle and store for max search
#            self.projs_matrix[i * K: (i + 1) *
#                K] = normaCoeffs * (self.inputa[0:K] * self.post_twid_vec).real
## self.projs_matrix[i , :] = normaCoeffs*(self.inputa[0:K]*
## self.post_twid_vec).real
#
##            # store new max score in tree
#            self.best_score_tree[i] = abs(self.projs_matrix[
#                i * K: (i + 1) * K]).max()

    # synthesizes the best atom through ifft computation (much faster than
    # closed form)
    def synthesize_atom(self, value=None):
        """ synthesizes the best atom through ifft computation (much faster than closed form)
            New version uses the PyWinServer to serve waveforms"""
        ###################" new version ############"
#        global _PyServer
#        print len(_PyServer.Waveforms)
        if value is None:
            return self.max_value * _PyServer.get_waveform(self.scale, self.max_bin_idx)
        else:
            return value * _PyServer.get_waveform(self.scale, self.max_bin_idx)
        ###################  old version ############
#        temp = zeros(2*self.scale)
#        temp[self.scale/2 + self.max_bin_idx] = self.max_value
#        waveform = zeros(2*self.scale)
# Number of frames : only 4 we need zeroes on the border before overlap-adding
##        P = 4
#        L = self.scale
#        K = L/2
#        y = zeros(L, complex)
#        x = zeros(L, complex)
#        # we're note in Matlab anymore, loops are more straight than indexing
#        for i in range(1,3):
#            y[0:K] = temp[i*K : (i+1)*K]
#
#            # do the pre-twiddle
#            y = y *  self.pre_i_twidVec
#
#            # compute ifft
#            x = ifft(y)
#
#            # do the post-twiddle
#            x = x * self.post_i_twidVec
#
#            x = 2*sqrt(1/float(L))*L*x.real*self.w_long
#
#            # overlapp - add
#            waveform[i*K : i*K +L] = waveform[i*K : i*K +L] +  x
#
#        # scrap zeroes on the borders
#        return waveform[L/2:-L/2]

    def plot_proj_matrix(self):
        import matplotlib.pyplot as plt
        plt.figure()
#        plt.subplot(211)
#        plt.plot(self.best_score_tree)
#        plt.subplot(212)
        plt.plot(self.projs_matrix)
        plt.title("Block-" + str(self.scale) + " best Score of" + str(self.max_value)
                  + " p :" + str(self.max_frame_idx) + " , k: " + str(self.max_bin_idx))

    def im_proj_matrix(self):
        import matplotlib.pyplot as plt
#        plt.figure()
        plt.imshow(np.abs(self.projs_matrix.reshape((self.frame_num, self.frame_len))),
                   aspect='auto',
                   interpolation='nearest')

class LOBlock(Block):
    """ Class that inherit classic MDCT block class and deals with local optimization
        This is the main class for differentiating LOMP from MP """

    # Typically all attributes are the same than mother class:
    maxTimeShift = 0
    adjustTimePos = True
    fftplan = None
    # constructor - initialize residual signal and projection matrix

    def __init__(self, length=0, resSignal=None, frameLen=0, tinvOptim=True, useC=True, forceHF=False):
        self.scale = length
        self.residual_signal = resSignal
        self.adjustTimePos = tinvOptim
        self.use_c_optim = useC
        self.HF = forceHF

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len

        # only difference , here, we keep the complex values for the cross
        # correlation
#        self.projs_matrix = zeros(len(self.framed_data_matrix) , complex)
        self.projs_matrix = np.zeros(len(self.framed_data_matrix), float)

# only the update method is interesting for us : we're hacking it to experiment
#    def update(self , newResidual , startFrameIdx=0 , stopFrameIdx=-1):
#        self.residual_signal = newResidual
#
#        if stopFrameIdx <0:
#            endFrameIdx = self.frameNumber -1
#        else:
#            endFrameIdx = stopFrameIdx
#        L = self.scale
#
#        # update residual signal
# self.framed_data_matrix[startFrameIdx*L/2 : endFrameIdx*L/2 + L] =
# self.residual_signal.dataVec[startFrameIdx*self.frameLength :
# endFrameIdx*self.frameLength + 2*self.frameLength]
#
#        # TODO changes here
##        self.computeMDCT(startFrameIdx , stopFrameIdx)
#        self.computeMCLT(startFrameIdx , stopFrameIdx)
#
#        # TODO changes here
#        self.find_max()
    # inner product computation through MDCT
    def compute_transform(self, startingFrame=1, endFrame=-1):
        if self.w_long is None:
            self.initialize()

        # due to later time-shift optimizations , need to ensure nothing is
        # selected too close to the borders!!

        if endFrame < 0 or endFrame > self.frame_num - 3:
            endFrame = self.frame_num - 3

        if startingFrame < 2:
            startingFrame = 2
        # new version: C binding
#        if self.use_c_optim:
        parallelProjections.project_mclt(self.framed_data_matrix, self.best_score_tree,
                                             self.projs_matrix,
                                             self.locCoeff,
                                             self.post_twid_vec,
                                             startingFrame,
                                             endFrame,
                                             self.scale)
#        else:
#            L = self.scale
#            K = L / 2
#            T = K / 2
#            normaCoeffs = sqrt(2 / float(K))
#
#            locenframedDataMat = self.framed_data_matrix
##            locfftMat = self.fftMat
##            locprojMat = self.projs_matrix
#
#            preTwidCoeff = self.locCoeff
#            postTwidCoeff = self.post_twid_vec
#
#            # Bottleneck here !! need to fasten this loop : do it by Matrix
#            # technique? or bind to a C++ file?
#            for i in range(startingFrame, endFrame):
#                x = locenframedDataMat[i * K - T: i * K + L - T]
#                if len(x) != L:
#                    x = np.zeros(L, complex)
#
#                # do the pre-twiddle
#                x = x * preTwidCoeff
#
#                # compute fft
##                locfftMat[: , i] = fft(x , L)
#
#                # post-twiddle
##                y = locfftMat[0:K , i] * postTwidCoeff
#                y = (fft(x, L)[0:K]) * postTwidCoeff
#    #            y = self.doPretwid(locfftMat[0:K , i], postTwidCoeff)
#
#                # we work with MCLT now
#                self.projs_matrix[i * K: (i + 1) * K] = normaCoeffs * y
#
#                # store new max score in tree
#                self.best_score_tree[i] = abs(self.
#                    projs_matrix[i * K: (i + 1) * K]).max()

#        if self.HF:
#            for i in range(startingFrame , endFrame):
# self.bestScoreHFTree[i] =
# abs(self.projs_matrix[(i+self.HFlimit)*self.scale/2 :
# (i+1)*self.scale/2]).max()
#

    # construct the atom that best correlates with the signal
    def get_max_atom(self, debug=0):

        self.max_frame_idx = floor(self.maxIdx / (0.5 * self.scale))
        self.max_bin_idx = self.maxIdx - self.max_frame_idx * (0.5 * self.scale)

        # hack here : let us project the atom waveform on the neighbouring
        # signal in the FFt domain,
        # so that we can find the maximum correlation and best adapt the time-
        # shift
        Atom = atom.Atom(self.scale, 1, max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0), self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx

        # re-compute the atom amplitude for IMDCT
#        if self.max_value.real < 0:
#            self.max_value = -abs(self.max_value)
#        else:
#            self.max_value = abs(self.max_value)

        Atom.mdct_value = self.max_value
        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom(value=1)
        Atom.time_shift = 0
        Atom.proj_score = 0.0

        input1 = self.framed_data_matrix[(self.max_frame_idx - 1.5) *
             self.scale / 2: (self.max_frame_idx + 2.5) * self.scale / 2]
        input2 = np.concatenate((np.concatenate(
            (np.zeros(self.scale / 2), Atom.waveform)), np.zeros(self.scale / 2)))

        # debug cases: sometimes when lot of energy on the borders , pre-echo
        # artifacts tends
        # to appears on the border and can lead to seg fault if not monitored
        # therefore we need to prevent any time shift leading to positions
        # outside of original signals
        # ie in the first and last frames.
#        if debug>0:
#            print self.max_frame_idx , self.maxIdx , self.frameNumber
#            if self.max_frame_idx ==1:
#                plt.figure()
#                plt.plot(Atom.waveform)
# plt.plot(self.framed_data_matrix[0 : Atom.timePosition + Atom.length],'r')
#                plt.show()
#
# print (self.max_frame_idx-1.5)  * self.scale/2 , (self.max_frame_idx+2.5)  *
# self.scale/2
#            print len(input1) , len(input2)
        if len(input1) != len(input2):
            print self.max_frame_idx, self.maxIdx, self.frame_num
            print len(input1), len(input2)
            #if debug>0:
            print "atom in the borders , no timeShift calculated"
            return Atom

        # retrieve additional timeShift
#        if self.use_c_optim:
        scoreVec = np.array([0.0])
        Atom.time_shift = parallelProjections.project_atom(
            input1, input2, scoreVec, self.scale)

        if abs(Atom.time_shift) > Atom.length / 2:
            print "out of limits: found time shift of", Atom.time_shift
            Atom.time_shift = 0
            return Atom

        self.maxTimeShift = Atom.time_shift
        Atom.time_position += Atom.time_shift

        # retrieve newly projected waveform
        Atom.proj_score = scoreVec[0]
        Atom.waveform *= Atom.proj_score
#            Atom.waveform = input2[self.scale/2:-self.scale/2]
#        else:
#            sigFft = fft(self.framed_data_matrix[(self.max_frame_idx - 1.5) * self.scale /
#                2: (self.max_frame_idx + 2.5) * self.scale / 2], 2 * self.scale)
#            atomFft = fft(np.concatenate((np.concatenate((np.zeros(self.scale / 2),
#                 Atom.waveform)), np.zeros(self.scale / 2))), 2 * self.scale)
#
#            Atom.time_shift, score = Xcorr.GetMaxXCorr(
#                atomFft, sigFft, maxlag=self.scale / 2)
#            self.maxTimeShift = Atom.time_shift
#            Atom.proj_score = score
#    # print "found correlation max of " ,
#    # float(score)/sqrt(2/float(self.scale))
#
#            # CAses That might happen: time shift result in choosing another
#            # atom instead
#            if abs(Atom.time_shift) > Atom.length / 2:
#                print "out of limits: found time shift of", Atom.time_shift
#                Atom.time_shift = 0
#                return Atom
#
#            Atom.time_position += Atom.time_shift
#
#            # now let us re-project the atom on the signal to adjust it's
#            # energy: Only if no pathological case
#
#            # TODO optimization : pre-compute energy (better: find closed form)
#            if score < 0:
#                Atom.amplitude = -sqrt(-score)
#                Atom.waveform = (
#                    -sqrt(-score / sum(Atom.waveform ** 2))) * Atom.waveform
#            else:
#                Atom.amplitude = sqrt(score)
#                Atom.waveform = (
#                    sqrt(score / sum(Atom.waveform ** 2))) * Atom.waveform
# projOrtho = sum(Atom.waveform * self.residual_signal.dataVec[Atom.timePosition
# : Atom.timePosition + Atom.length])

#        if score <0:
#            Atom.amplitude = -1
#            Atom.waveform = -Atom.waveform

        return Atom

    def plot_proj_matrix(self):
        import matplotlib.pyplot as plt
        maxFrameIdx = floor(self.maxIdx / (0.5 * self.scale))
        maxBinIdx = self.maxIdx - maxFrameIdx * (0.5 * self.scale)
        plt.figure()
#        plt.subplot(211)
#        plt.plot(self.best_score_tree)
#        plt.subplot(212)
        plt.plot(abs(self.projs_matrix))
        plt.title("Block-" + str(self.scale) + " best Score of" + str(abs(self.max_value)) + " at " + str(self.maxIdx)
                  + " p :" + str(maxFrameIdx)
                  + " , k: " + str(maxBinIdx)
                  + " , l: " + str(self.maxTimeShift))


class FullBlock(Block):
    """ Class that inherit classic MDCT block class and but contains all time localizations """
    # parameters
    maxKidx = 0
    maxLidx = 0

    # constructor - initialize residual signal and projection matrix
    def __init__(self, length=0, resSignal=None, frameLen=0):
        self.scale = length
        self.residual_signal = resSignal

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len

        # ok here the mdct will be computed for every possible time shift
        # so allocate enough memory for all the transforms

        # here the projection matrix is actually a list of K by K matrices
        self.projs_matrix = dict()
        self.best_score_tree = dict()

        # initialize an empty matrix of size len(data) for each shift
        for i in range(self.scale / 2):
#            self.projs_matrix[i] = zeros((self.scale/2 , self.scale/2))
            self.projs_matrix[i] = np.zeros(len(self.framed_data_matrix))
            self.best_score_tree[i] = np.zeros(self.frame_num)

    def initialize(self):

        #Windowing
        L = self.scale

        self.w_long = np.array([np.sin(float(l + 0.5) * (np.pi / L)) for l in range(L)])

        # twidlle coefficients
        self.pre_twid_vec = np.array([exp(n * (-1j) * np.pi / L) for n in range(L)])
        self.post_twid_vec = np.array(
            [exp((float(n) + 0.5) * -1j * np.pi * (L / 2 + 1) / L) for n in range(L / 2)])

        if self.windowType == 'half1':
            self.w_long[0:L / 2] = 0
            # twidlle coefficients
            self.pre_twid_vec[0:L / 2] = 0

        # OPTIM -> do pre-twid directly in the windows
        self.locCoeff = self.w_long * self.pre_twid_vec

    # The update method is nearly the same as CCBlock
    def update(self, new_res_signal, startFrameIdx=0, stopFrameIdx=-1):
        self.residual_signal = new_res_signal

        if stopFrameIdx < 0:
            endFrameIdx = self.frame_num - 1
        else:
            endFrameIdx = stopFrameIdx

        # debug : recompute absolutely all the products all the time

        L = self.scale

# print "Updating Block " , L , " from frame " , startFrameIdx , " to " ,
# endFrameIdx

        # update residual signal
        self.framed_data_matrix[startFrameIdx * L / 2: endFrameIdx * L / 2 + L] = self.residual_signal.data[startFrameIdx * self.frame_len: endFrameIdx * self.frame_len + 2 * self.frame_len]

        # TODO changes here
        self.compute_transform(startFrameIdx, stopFrameIdx)

        # TODO changes here
        self.find_max()

    # inner product computation through MCLT with all possible time shifts
    def compute_transform(self, startingFrame=1, endFrame=-1):
        if self.w_long is None:
            self.initialize()

        if endFrame < 0:
            endFrame = self.frame_num - 1

        if startingFrame < 1:
            startingFrame = 1

        startingFrame = 1
        endFrame = self.frame_num - 1

        L = self.scale
        K = L / 2
#        T = K/2
#        normaCoeffs = sqrt(2/float(K))
#        print startingFrame , endFrame
        for l in range(-K / 2, K / 2, 1):
            parallelProjections.project(self.framed_data_matrix,
                                                 self.best_score_tree[l + K / 2],
                                                 self.projs_matrix[
                                                     l + K / 2],
                                                 self.locCoeff,
                                                 self.post_twid_vec,
                                                 startingFrame,
                                                 endFrame,
                                                 self.scale, l)

    def find_max(self):
        K = self.scale / 2
        bestL = 0
        treeMaxIdx = 0
        bestSCore = 0
        for l in range(-K / 2, K / 2, 1):
            if self.best_score_tree[l + K / 2].max() > bestSCore:
                bestSCore = self.best_score_tree[l + K / 2].max()
                treeMaxIdx = self.best_score_tree[l + K / 2].argmax()
                bestL = l

        maxIdx = abs(self.projs_matrix[bestL + K / 2]).argmax()

        self.maxLidx = bestL

        self.maxIdx = maxIdx
        self.max_frame_idx = treeMaxIdx
        self.max_value = self.projs_matrix[bestL + K / 2][maxIdx]

#        print "Max Atom : " , self.maxIdx , self.maxLidx , self.max_value
    # construct the atom that best correlates with the signal
    def get_max_atom(self):
        self.max_bin_idx = self.maxIdx - self.max_frame_idx * (0.5 * self.scale)

        Atom = atom.Atom(self.scale, 1, max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0), self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx

        # re-compute the atom amplitude for IMDCT
        Atom.mdct_value = self.max_value

        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom()

        Atom.time_shift = -self.maxLidx  # + self.scale/4
        self.maxTimeShift = Atom.time_shift

        Atom.time_position += Atom.time_shift

        return Atom

    def plot_proj_matrix(self):
        ''' For debug purposes only... '''
        import matplotlib.pyplot as plt
        plt.figure()
#        plt.subplot(211)
        plt.plot(abs(self.projs_matrix[self.maxLidx + self.scale / 4]))
#        plt.title("reference l=0 , best k of "
#                  + str(bestK)
# + " max score of : " +
# str(abs(self.projs_matrix[self.maxIdx][bestK,:]).max())
# + " at l: " + str(abs(self.projs_matrix[self.maxIdx][bestK,:]).argmax() -
# self.scale/4) )
#        plt.subplot(212)
#        plt.imshow(abs(self.projs_matrix[self.maxIdx]) ,
#                   aspect='auto', cmap = cm.get_cmap('greys') ,
#                   extent=(-self.scale/4 , self.scale/4, self.scale/2, 0) ,
#                   interpolation = 'nearest')
        plt.title("Block-" + str(self.scale) + " best Score of" + str(self.max_value)
                  + " p :" + str(self.max_frame_idx)
#                  +" , k: "+ str(self.maxKidx)
                  + " , l: " + str(self.maxLidx))


class SpreadBlock(Block):
    ''' Spread MP is a technique in which you penalize the selection of atoms near existing ones
        in a  predefined number of iterations, or you specify a perceptual TF masking to enforce the selection of
        different features. The aim is to have maybe a loss in compressibility in the first iteration but
        to have the most characteristics / discriminant atoms be chosen in the first iterations '''

    # parameters
    distance = None
    mask = None
    maskSize = None
    penalty = None

    def __init__(self, length=0, resSignal=None, frameLen=0, useC=True, forceHF=False,
                 debug_level=None, penalty=0.5, maskSize=1):

        if debug_level is not None:
            _Logger.set_level(debug_level)

        self.scale = length
        self.residual_signal = resSignal

        if frameLen == 0:
            self.frame_len = length / 2
        else:
            self.frame_len = frameLen
        if self.residual_signal == None:
            raise ValueError("no signal given")

        self.framed_data_matrix = self.residual_signal.data
        self.frame_num = len(self.framed_data_matrix) / self.frame_len
        
        self.projs_matrix = np.zeros((self.frame_num* self.frame_len,))
        
        
        self.use_c_optim = useC
        self.HF = forceHF
        _Logger.info('new MDCT block constructed size : ' + str(self.scale))

        self.penalty = penalty
        # initialize the mask: so far no penalty
        self.mask_index_x = np.array([], dtype=int, ndmin=1)
        self.mask_index_y = np.array([], dtype=int, ndmin=1)
        
        self.mask = np.ma.empty_like(self.projs_matrix)
        self.mask_values = np.ma.ones(self.projs_matrix.shape)
        
        self.mask_index = np.zeros(self.projs_matrix.shape)
        self.maskSize = maskSize

    def find_max(self, it=-1):
        ''' Apply the mask to the projection before choosing the maximum '''

        # cannot use the tree indexing any more... too bad
#        treeMaxIdx = self.best_score_tree.argmax()
#        print self.mask_index_x, self.mask_index_y
#        indices = (self.mask_index_x * self.scale/2) + self.mask_index_y
#        for z in self.mask_index_x:            
#            self.projs_matrix[(z*self.scale/2) + self.mask_index_y] *= self.mask[(z*self.scale/2) + self.mask_index_y]
#        print self.framed_data_matrix.shape, self.mask.shape
#        np.ma.masked_equal(self.mask, 1)

#        self.mask_index = (self.mask < 1)

#        newImage = np.where(np.logical_and(image1, image2), (image1 + image2) / 2, 0)

#        self.projs_matrix = np.where(np.logical_and(self.projs_matrix, self.mask), self.projs_matrix * self.mask, self.projs_matrix)        
#        print self.projs_matrix.shape
#        print self.mask.mask
        np.ma.masked_equal(self.mask_values, 1)
        self.projs_matrix *= self.mask_values
        
#        for x,y in zip(self.mask_index_x, self.mask_index_y):
#            self.projs_matrix[x*self.scale/2 + y] *= self.mask[x*self.scale/2 + y]
#        plt.figure()
#        plt.imshow(reshape(self.mask,(self.frameNumber,self.scale/2)),interpolation='nearest',aspect='auto')
        self.maxIdx = np.abs(self.projs_matrix).argmax()

#        print self.maxIdx

# maxIdx = abs(self.projs_matrix[treeMaxIdx*self.scale/2 :
# (treeMaxIdx+1)*self.scale/2]).argmax()
#        self.maxIdx = maxIdx + treeMaxIdx*self.scale/2
        self.max_value = self.projs_matrix[self.maxIdx]

    def get_max_atom(self, HF=False):
        self.max_frame_idx = floor(self.maxIdx / (0.5 * self.scale))
        self.max_bin_idx = self.maxIdx - self.max_frame_idx * (0.5 * self.scale)
     
        # proceed as usual
        Atom = atom.Atom(self.scale, 1, max((self.max_frame_idx * self.scale / 2) - self.scale / 4, 0), self.max_bin_idx, self.residual_signal.fs)
        Atom.frame = self.max_frame_idx

        Atom.mdct_value = self.max_value

        # new version : compute also its waveform through inverse MDCT
        Atom.waveform = self.synthesize_atom()

        return Atom

    def compute_mask(self, best_atom, mask_bin_width, mask_frame_width, penalty):
        ''' update the mask, centering of the best current atom
        with specified Frequency and time width ''' 
        # update the mask : penalize the choice of an atom overlapping in time
        # and or frequency
        c_frame = int(np.ceil(best_atom.time_position / (self.scale / 2)))
        c_bin = int(best_atom.reduced_frequency * self.scale )        
        
#        print best_atom
#        print "Evaluating: ", self.scale, c_frame, c_bin
        
#        z1 = np.arange(int(c_frame - mask_frame_width), int(c_frame + mask_frame_width) +1,1)
#        z2 = np.arange(int(c_bin - mask_bin_width), int(c_bin + mask_bin_width) + 1,1)
#        x, y = np.meshgrid(z1, z2)
#        print self.mask_index_x, x
#        print x.shape 
#        self.mask_index_x = np.concatenate((self.mask_index_x, z1))
#        self.mask_index_y = np.concatenate((self.mask_index_y, z2))
#        for z in z1:
#            self.mask[(z*self.scale/2) + z2] *= penalty
#        self.mask_index.append()
        for i in range(-mask_frame_width, mask_frame_width + 1, 1):
#            self.mask_index.append()
#            self.mask_index[((c_frame + i) * self.scale / 2) + (c_bin - mask_bin_width): ((c_frame + i) * self.scale / 2) + (c_bin + mask_bin_width)] = 1
            self.mask_values[((c_frame + i) * self.scale / 2) + (c_bin - mask_bin_width): ((c_frame + i) * self.scale / 2) + (c_bin + mask_bin_width)] *= self.penalty

        
#            self.mask[((self.max_frame_idx + i) * self.scale / 2) + (self.max_bin_idx - self.maskSize): ((self.max_frame_idx + i) * self.scale / 2) + (self.max_bin_idx + self.maskSize)] = self.penalty
#       


#*****************************************************************************/
#                                                                            */
#                               approx.py                                    */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */

import math
import numpy as np

from PyMP import signals
from PyMP import log
from PyMP.base import BaseAtom, BaseApprox
from PyMP.tools.mdct import imdct


# global _Logger
_Logger = log.Log('Approx')


class Approx(BaseApprox):
    """ The approx class encapsulate the approximation that is iteratively being constructed by a greed algorithm
    
    Attributes
    ----------
    *dico* : :class:`.BaseDico`
        the dictionary (as a :class:`.BaseDico` object) from which it has been constructed
    *atoms* : list
        a list of :class:`.BaseAtom` objets
    *atom_number* : int
        the length of the atoms list
    *srr* : float
        the Signal to Residual Ratio achieved by the approximation
    *original_signal* : :class:`.Signal`
        a :class:`.Signal` object that is the original signal
    *length* : int
        Length in samples of the time signal, same as the one of `original_signal`
    *recomposed_signal* :  :class:`.Signal`
        a :class:`.Signal` objet for the reconstructed signal (as the weighted sum of atoms specified in the atoms list)
     
     """

    # Attributes
    dico = []
    atoms = []
    atom_number = 0
    srr = 0
    original_signal = []
#    frame_num = 0
#    frame_len = 0
    length = 0
    recomposed_signal = None
    fs = 0

    def __init__(self, dico=None, atoms=[], original_signal=None, length=0,
                 Fs=0, debug_level=None, fast_create=False):
        """ 
        Parameters
        ----------
        dico :  :class:`.BaseDico`
            the dictionary (as a :class:`.BaseDico` object) from which it has been constructed
        *atoms* : list
            a list of :class:`.BaseAtom` objets
                         
             """
        if debug_level is not None:
            _Logger.set_level(debug_level)

        self.dico = dico
        self.atoms = atoms
        self.original_signal = original_signal
        if original_signal != None:
            self.length = original_signal.length
            self.fs = original_signal.fs
            isNormalized = original_signal.is_normalized
        else:
            self.length = length
            self.fs = Fs
            isNormalized = False
#        if dico != None:
#            self.frame_num = self.length / max(self.dico.sizes)
#        if self.frame_num > 0:
#            self.frame_len = self.length / self.frame_num

        # We need to create a signal that has the same nature as the one
        # provided
        if original_signal is not None:
            self.recomposed_signal = original_signal.__class__(
                np.zeros(original_signal.data.shape), self.fs,
                debug_level=debug_level,
                fast_create=fast_create)
        else:
            self.recomposed_signal = signals.Signal(
                np.zeros(self.length), self.fs,debug_level=debug_level,
                fast_create=fast_create)

        self.recomposed_signal.is_normalized = isNormalized

    def synthesize(self, method=0, forceReSynthesis=True):
        """ function that will synthesise the approximant using the list of atoms
            this is mostly DEPRECATED"""
        if self.original_signal == None:
            _Logger.warning("No original Signal provided")
#            return None

        if (self.recomposed_signal == None) | forceReSynthesis:
            synthesizedSignal = np.zeros(self.length)

            if len(self.atoms) == 0:
                _Logger.info("No Atoms")
                return None

            # first method by inverse MDCT
            if method == 0:
                for mdctSize in self.dico.sizes:
                    mdctVec = np.zeros(self.length)
                    for atom in self.atoms:
                        if atom.length == mdctSize:
                            # bugFIx
                            n = atom.time_position + 1
                            frame = math.floor(
                                float(n) / float(atom.length / 2)) + 1
# mdctVec[frame*float(atom.length /2) + atom.frequencyBin] += atom.amplitude
                            mdctVec[frame * float(atom.length /
                                                  2) + atom.freq_bin] += atom.mdct_value
                    synthesizedSignal += imdct(mdctVec, mdctSize)
# synthesizedSignal += concatenate((zeros(mdctSize/4) , imdct(mdctVec ,
# mdctSize)) )[1:-mdctSize/4+1]

            # second method by recursive atom synthesis - NOT WORKING
            elif method == 1:
                for atom in self.atoms:
                    atom.synthesize_ifft()
                    synthesizedSignal[atom.time_position:
                                      atom.time_position + atom.length] += atom.waveform

            # HACK here to resynthesize using LOMP atoms
            elif method == 2:
                for atom in self.atoms:
                    atom.waveForm = atom.synthesize_ifft()
                    if (atom.proj_score is not None):
                        if (atom.proj_score < 0):
                            atom.waveform = (-math.sqrt(-atom.proj_score /
                                                        sum(atom.waveform ** 2))) * atom.waveform
                        else:
                            atom.waveform = (math.sqrt(atom.proj_score /
                                                       sum(atom.waveform ** 2))) * atom.waveform

                    synthesizedSignal[atom.time_position:
                                      atom.time_position + atom.length] += atom.waveform

            self.recomposed_signal = signals.Signal(synthesizedSignal, self.fs)
            # return self.recomposed_signal
        # other case: we just give the existing synthesized Signal.
        return self.recomposed_signal

    def __getitem__(self, item):
        ''' Get the waveform built by only the subAtomNumber first atoms
            Outputs:

                *  a py_pursuit_Signal array '''
        if isinstance(item, slice):
            start, stop, step = item.start, item.stop, item.step
        elif isinstance(item, int):
            start = item
            stop = item
            return self.atoms[item]

        else:
            raise TypeError("not recognized")

        if start is None:
            start = 0
        
        if step is None:
            step = 1
        output_approx = Approx(
            self.dico, [], self.original_signal, self.length, self.fs)
        if stop > self.atom_number:
            raise ValueError('Dude you asked fore more than I can give you..')
                
        for atomIdx in range(start, stop, step):
            output_approx.add(self.atoms[atomIdx])

        return output_approx

    def __repr__(self):
        return 'Approx Object: %d atoms, SRR of %2.2f dB' % (self.atom_number, self.srr)

    def get_neighbors(self, atom):
        """ returns the atom neighbor indexes """
        t_interv = [atom.time_position - self.dico.get_pad(),
                    atom.time_position + atom.length]

        return [i for i in range(self.atom_number) if
                (self.atoms[i].time_position >= t_interv[0]) and
                (self.atoms[i].time_position < t_interv[1]) and
                (self.atoms[i].time_position + self.atoms[i].length > atom.time_position)]

    # Filter the atom list by the given criterion
    def filter_atoms(self, scale=0, time_interv=None, freq_interv=None):
        '''Filter the atom list by the given criterion, returns an new approximant object'''
        filteredApprox = Approx(self.dico, [], self.original_signal,
                                self.length, self.fs)

        # TODO: change to a list sorting: will be much faster
        for atom in self.atoms:
            doAppend = True
#            if atom.length == scale:
##                print  atom.length , " appended "
#                filteredApprox.add(atom)

            if atom.length == scale:
                doAppend &= True
            else:
                doAppend &= False

            if(time_interv != None):
                if (time_interv[0] < atom.time_position) and (atom.time_position <= time_interv[1]):
                    doAppend = True
                else:
                    doAppend = False

            if(freq_interv != None):
                if (min(freq_interv) < atom.reduced_frequency <= max(freq_interv)):
                    doAppend &= True
                else:
                    doAppend &= False
            if doAppend:
#                print  atom.length , " appended "
                filteredApprox.add(atom)

        return filteredApprox

    # this function adds an atom to the collection and updates the internal
    # signal approximant
    def add(self, newAtom, clean=False, noWf=False):
        '''this function adds an atom to the collection and updates
        the internal signal approximant'''
        if not isinstance(newAtom, BaseAtom):
            raise TypeError("add need a py_pursuit_Atom as parameter ")

        self.atoms.append(newAtom)
        self.atom_number += 1
        # add atom waveform to the internal signal if exists
        if not noWf:
            if self.recomposed_signal != None:
                self.recomposed_signal.add(newAtom)
            else:
                self.synthesize(0, True)
    #            print "approx resynthesized"
            if clean:
                del newAtom.waveform

        # otherwise recompute ?


    def update(self, atom_indexes, new_weights, update_rec=False):
        """ update atom values (e.g. after projection) and the
            recomposed_signal instance

            """
        if update_rec:
            for i in range(len(atom_indexes)):
                ind = atom_indexes[i]
                atom = self.atoms[ind]
                self.remove(atom, position=ind)
                atom.mdct_value = new_weights[i]

                self.add(atom)
        else:
            for i in range(len(atom_indexes)):
                self.atoms[atom_indexes[i]].mdct_value = new_weights[i]

    def remove(self, atom, position=None):
        ''' We need a routine to remove an atom , by default the last atom is removed '''
        if not isinstance(atom, BaseAtom):
            raise TypeError("add need a BaseAtom as parameter ")
        if position is None:
            self.atoms.remove(atom)
        else:
            self.atoms.pop(position)
        self.atom_number -= 1

        if self.recomposed_signal != None:
            self.recomposed_signal.subtract(atom, prevent_energy_increase=False)

    # routine to compute approximation Signal To Residual ratio so far
    def compute_srr(self, residual=None):
        ''' routine to compute approximation Signal To Residual ratio so far using:

            .. math::srr = 10 \log_10 \frac{\| \tilde{x} \|^2}{\| \tilde{x} - x \|^2}

            where :math:`\tilde{x}` is the reconstructed signal and :math:`x` the original
        '''
        if not isinstance(self.recomposed_signal, signals.Signal):
            return np.NINF

#        recomposedEnergy = sum((self.recomposed_signal.dataVec)**2)
        recomposedEnergy = self.recomposed_signal.energy

        if recomposedEnergy <= 0:
            return np.NINF

        if residual is None:
            resEnergy = np.sum((self.original_signal.data -
                                self.recomposed_signal.data) ** 2)
        else:
            resEnergy = residual.energy
# resEnergy = sum((self.original_signal.dataVec -
# self.recomposed_signal.dataVec)**2)
        if resEnergy == 0:
            return np.PINF

        self.srr = 10 * math.log10(recomposedEnergy / resEnergy)
        return self.srr

    # ploting routine 2d in a time-frequency plane
    def plot_tf(
        self, labelX=True, labelY=True, fontsize=12, ylim=None, patchColor=None, labelColor=None,
        multicolor=False, axes=None, maxAtoms=None, recenter=None, keepValues=False,
            french=False, Alpha=False, logF=False, cmap=None):
        """ A time Frequency plot routine using Matplotlib

            each atom of the approx is plotted in a time-frequency plane with a gray scale for amplitudes
            Many options are available:

            - labelX     : whether or not to add the Time label on the X axis

            - labelY     : whether or not to add the Frequency label on the Y axis

            - fontsize    : the label fontSize

            - ylim        : the frequency range of the plot

            - patchColor    : specify a single color for all atoms

            - maxAtoms    : specify a maximum number of atom to take into account. Usually atoms are ordered by decreasing amplitude due to MP rules, meaning only the most salient atoms will be plotted

            - recenter    : a couple of values to offset the time and frequency localization of atoms

            - keepValues    : use atom Amplitudes for atom colorations

            - french    : all labels in french

            - Alpha     : use transparency

            - logF        : frequency axis in logarithmic scale

        """
        import matplotlib.patches as mpatches
        import matplotlib
        import matplotlib.pyplot as plt
        from matplotlib.collections import PatchCollection
        import matplotlib.colors as cc
#        plt.figure()
        if french:
            labx = "Temps (s)"
            laby = "Frequence (Hz)"
        else:
            labx = "Time (s)"
            laby = "Frequency (Hz)"

        Fs = self.fs
        # first loop to recover max value
        maxValue = 0
        maxFreq = 1.0
        minFreq = 22000.0
        if not keepValues:
            valueArray = np.array(
                [math.log10(abs(atom.get_value())) for atom in self.atoms])
        else:
            valueArray = np.array(
                [abs(atom.get_value()) for atom in self.atoms])
        if multicolor:
            if cmap is None:
                cmap = cc.LinearSegmentedColormap('jet', abs(valueArray))
            else:
                cmap = cc.LinearSegmentedColormap(cmap, abs(valueArray))
            for atom in self.atoms:
                if abs(atom.mdct_value) > maxValue:
                    maxValue = abs(atom.mdct_value)

        # normalize to 0 -> 1
#        if min(valueArray) < 0:
        if len(valueArray) > 0 and not keepValues:
            valueArray = valueArray - min(valueArray)
            valueArray = valueArray / max(valueArray)

        # Get target axes
        if axes is None:
            axes = plt.gca()

        # normalize the colors
#        if multicolor:
#            normedValues = cc.Normalize(valueArray)

        if maxAtoms is not None:
            maxTom = min(maxAtoms, len(self.atoms))
        else:
            maxTom = len(self.atoms)

        # Deal with static offsets
        if recenter is not None:
            yOffset = recenter[0]
            xOffset = recenter[1]
        else:
            yOffset, xOffset = 0, 0

#        print yOffset,xOffset

        if Alpha:
            alphaCoeff = 0.6
        else:
            alphaCoeff = 1
        patches = []
        colors = []
        for i in range(maxTom):
            atom = self.atoms[i]
            f, bw, p, l = atom.get_tf_info(Fs)
#            print "f =  " ,f , " Hz"
            if f > maxFreq:
                maxFreq = f + bw + bw
            if f < minFreq:
                minFreq = f - bw - 10 * bw

#            xy = p, f-bw,
            xy = p - xOffset, f - yOffset,
            width, height = l, bw
#            xy = p - xOffset, log10(f) - yOffset,

            if logF:
                xy = p - xOffset, (12 * np.log2(f - yOffset))

                height = 1.0  # BUGFIX

#            print xy , f
            colorvalue = math.sqrt(valueArray[i])

            if multicolor:
                patchtuplecolor = (math.sqrt(colorvalue),
                                   math.exp(
                                       -((colorvalue - 0.35) ** 2) / (0.15)),
                                   1 - math.sqrt(colorvalue))
                colors.append(colorvalue)
            else:
                if patchColor is None:
                    patchtuplecolor = (1 - math.sqrt(colorvalue),
                                       1 - math.sqrt(colorvalue),
                                       1 - math.sqrt(colorvalue))
                    colors.append(colorvalue)
                else:
                    patchtuplecolor = patchColor
                    colors.append(patchColor)
#            patchtuplecolor = (colorvalue, colorvalue, colorvalue)
# patch = mpatches.Rectangle(xy, width, height, facecolor=patchtuplecolor,
# edgecolor=patchtuplecolor , zorder=colorvalue )
#            xy = p + L/2 , f-bw,
            patch = mpatches.Ellipse(
                xy, width, height, facecolor=patchtuplecolor,
                edgecolor=patchtuplecolor, zorder=colorvalue)
            if keepValues:
                patches.append(patch)
            else:
                axes.add_patch(patch)

        if keepValues:

            # first sort the patches by values
            sortedIndexes = valueArray.argsort()
            PatchArray = np.array(patches)
            p = PatchCollection(PatchArray[sortedIndexes].tolist(),
                                linewidths=0.,
                                cmap=matplotlib.cm.copper_r,
                                match_original=False,
                                alpha=alphaCoeff)
    #        print array(colors).shape
            p.set_array(valueArray[sortedIndexes])

            axes.add_collection(p)

        if labelColor is None:
            labelColor = 'k'

        # finalize graphic options
        axes.set_xlim(0 - xOffset, (self.length) / float(Fs) - xOffset)
        if ylim is None:
            axes.set_ylim(max(0, minFreq) - yOffset, maxFreq - yOffset)
        else:
            axes.set_ylim(ylim)
        if labelY:
            axes.set_ylabel(laby, fontsize=fontsize, color=labelColor)
#            plt.ylabel.set_color(labelColor)
        else:
            plt.setp(axes, 'yticklabels', [])

        if labelX:
            axes.set_xlabel(labx, fontsize=fontsize)
        else:
            plt.setp(axes, 'xticklabels', [])
        if keepValues:
            plt.colorbar(p)

    def plot_3d(self, itStep, fig=None):
        """ Creates a 3D Time-Frequency plot with various steps
        NOT WORKING below 0.99
        EXPERIMENTAL"""
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        if fig is None:
            fig = plt.figure()

        # beware of compatibility issues here
        ax = Axes3D(fig)
        # for atom, idx in zip(self.atoms, range(self.atom_number)):
        ts = [atom.time_position / float(self.fs)
              for atom in self.atoms]
        fs = [atom.reduced_frequency * self.
              fs for atom in self.atoms]
        zs = range(self.atom_number)
# ss = [float(atom.length)/float(self.fs) for atom in
# self.atoms]

        ax.scatter(ts, zs, fs, color="k", s=10, marker='o')
        ax.set_xlabel('Temps (s)')
        ax.set_ylabel('Iteration ')
        ax.set_zlabel('Frequence (Hz) ')

    def dump(self, dumpFilePath):
        import cPickle

        output = open(dumpFilePath, 'wb')
        _Logger.info("Dumping approx to  " + dumpFilePath)
        cPickle.dump(self, output, protocol=0)
        _Logger.info("Done ")

    def to_dico(self):
        """ Returns the approximant as a sparse dictionary object ,
        key is the index of the atom and values are atom objects"""
        dico = {}

        for atom in self.atoms:
            block = [i for i in range(
                len(self.dico.sizes)) if self.dico.sizes[i] == atom.length][0]
            n = atom.time_position + 1
            frame = math.floor(float(n) / float(atom.length / 2)) + 1
# sparseArray[int(block*self.length +  frame*float(atom.length /2) +
# atom.frequencyBin)] = (atom.mdct_value , frame*float(atom.length /2) -
# atom.timePosition)
# dico[int(block*self.length +  frame*float(atom.length /2) +
# atom.frequencyBin)] = atom
            key = int(block * self.length + frame * float(atom.
                                                          length / 2) + atom.freq_bin)
            if key in dico:
                dico[key].append(atom)
            else:
                dico[key] = [atom]

        return dico

    def to_sparse_array(self):
        """ Returns the approximant as a sparse dictionary object , key is the index of the atom and value is its amplitude"""
        sparseArray = {}
        # quickly creates a dictionary for block numbering
        blockIndexes = {}
        for i in range(len(self.dico.sizes)):
            blockIndexes[self.dico.sizes[i]] = i

        for atom in self.atoms:
            block = blockIndexes[atom.length]
            n = atom.time_position + 1
            frame = math.floor(float(n) / float(atom.length / 2)) + 1
# sparseArray[int(block*self.length +  frame*float(atom.length /2) +
# atom.frequencyBin)] = (atom.mdct_value , frame*float(atom.length /2) -
# atom.timePosition)
            sparseArray[int(block * self.length + frame * float(atom.length / 2)
                            + atom.freq_bin)] = (atom.mdct_value, atom.time_position)
        return sparseArray

#    def toSparseMatrix(self):
# """ Returns the approximant as a sparse dictionary object , key is the index
# of the atom and value is its amplitude"""
#        sparseArray = {}
#        # quickly creates a dictionary for block numbering
#        blockIndexes = {}
#        for i in range(len(self.dico.sizes)):
#            blockIndexes[self.dico.sizes[i]] = i
#
#        for atom in self.atoms:
#            block = blockIndexes[atom.length]
#            n = atom.timePosition +1
#            frame = math.floor( float(n) / float(atom.length /2) ) +1
# sparseArray[int(block*self.length +  frame*float(atom.length /2) +
# atom.frequencyBin)] = (atom.mdct_value , frame*float(atom.length /2) -
# atom.timePosition)
# sparseArray[int(block*self.length +  frame*float(atom.length /2) +
# atom.frequencyBin)] = (atom.mdct_value , atom.timePosition)
#        return sparseArray

    def to_array(self):
        """ Returns the approximant as an array object , key is the index of the atom and value is its amplitude"""
        sparseArray = np.zeros(len(self.dico.sizes) * self.length)
        timeShifts = np.zeros(len(self.dico.sizes) * self.length)
        # quickly creates a dictionary for block numbering
        blockIndexes = {}
        for i in range(len(self.dico.sizes)):
            blockIndexes[self.dico.sizes[i]] = i

        for atom in self.atoms:
            block = blockIndexes[atom.length]
            n = atom.time_position + 1
            frame = math.floor(float(n) / float(atom.length / 2)) + 1
            sparseArray[int(block * self.length + frame * float(
                atom.length / 2) + atom.freq_bin)] = atom.mdct_value
            timeShifts[int(block * self.length + frame * float(atom.length / 2) + atom.freq_bin)] = frame * float(atom.length / 2) - atom.time_position - atom.length / 4
        return sparseArray, timeShifts

#    def __del__(self):
#        """ Full Cleaning is sometimes useful """
#        for atom in self.atoms:
#            del atom.waveform
#            del atom
#        del self.recomposed_signal.dataVec


def load(inputFilePath):
    import cPickle
    file_obj = open(inputFilePath, 'r')
    return cPickle.load(file_obj)


def read_from_mat_struct(matl_struct):
    """ retrieve the python object from a saved version in matlab format
    This is useful if you saved an approx object using the scipy.io.savemat
    routine and you loaded it back using scipy.io.loadmat"""

    # Convert to object struct
    appObject = matl_struct[0, 0]
    dico = appObject.dico[0, 0]
    # TODO proper lecture of dictionary objects

    # So far we only read the parameters

    approx = Approx(None, [], None, appObject.length[0, 0], appObject.
                    fs[0, 0])
    return approx


def fusion_approxs(approxCollection, unPad=True):
    """ fusion a collection of frame by frame approximation.
    The collection is assumed to be temporally ordered """

    if not len(approxCollection) > 0:
        raise ValueError('Approx collection appears to be actually empty')

    dico = approxCollection[0].dico
    approxLength = approxCollection[0].length * len(approxCollection)
    if unPad:
        approxLength -= 2 * dico.sizes[-1] * len(approxCollection)
    Fs = approxCollection[0].fs

    approx = Approx(dico, [], None, approxLength, Fs)
    for segIdx in range(len(approxCollection)):
        currentApprox = approxCollection[segIdx]
        offset = segIdx * currentApprox.length
        if unPad:
            offset -= (segIdx + 1) * dico.sizes[-1]

        for atom in currentApprox.atoms:
            atom.time_position += offset
            approx.atoms.append(atom)

    return approx



#                                                                            */
#                               win_server.py                                */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */
#
'''
 This class avoid the recomputation of the atoms by ensuring waveforms are
 just computed once then re-used

 TO BE REFACTORED
                                                                       */
'''

import numpy as np
import cmath
import math
# import fftw3
#from numpy.fft import ifft
from scipy.signal import hann

# IMPORTANT: this class requires that the Fast Projection C library is
# compiled and accessible
try:
    import parallelProjections
except ImportError:
    print '''Failed to load C fast projection library, please make sure it is
            compiled and/or accessible
            Run test_Cmodule.py for more information'''
    raise ImportError('No module named parallelProjections')


class PyServer():
    """ Musical Signal and audio in general happens to be well described by a
    collection of atoms of variable sizes, frequency and time localizations.
    However, due their redundancy results in the fact that often,
    the same waveforms are used at different time locations. This being said,
    one might hope substantial improvement by avoiding to recompute these
    waveform each tile ther're needed! """

    type = 'MDCT'  # TODO so far we only create MDCT atoms with sinus windows
    Waveforms = {}
        # Waveforms are stored in a dictionary, the key is Length+Frequency

    PreTwidCoeffs = {}
        # dictionary of pre-twidlle coefficient: key is the scale
    PostTwidCoeff = {}
        # dictionary of post-twidlle coefficient: key is the scale
    Windows = {}  # Sine Windows : key is the scale

    __single = None

    def __init__(self, useC=True):
        """ The Server is global"""
        if PyServer.__single:
            raise PyServer.__single
        PyServer.__single = self
        self.use_c_optim = useC

    def get_waveform(self, scale, binIndex):
        ''' Check whether the waveform is already in the dictionary
            If waveform has already been created, no need to recompute it'''
        key = scale + binIndex
        if not key in self.Waveforms:
            self.Waveforms[key] = self.create_waveform(scale, binIndex)

        return self.Waveforms[key]

    def create_waveform(self, scale, binIndex):
        ''' By default, will use the C library to create the waveform'''
        if self.use_c_optim:
            return parallelProjections.get_atom(int(scale), int(binIndex))


def get_server(x=PyServer):
    """ You should only access servers via this class method:
        if no instance is available, it will create one, otherwise
        it will return the singleton """
    try:
        single = x()
    except PyServer, s:
        single = s
    return single


class PyGaborServer():
    """ Musical Signal and audio in general happens to be well described by a collection of atoms
        of variable sizes, frequency and time localizations. However, due their redundancy results
        in the fact that often, the same waveforms are used at different time locations. This being said,
        one might hope substantial improvement by avoiding to recompute these waveform each tile ther're
        needed! """

    type = 'Gabor'  # TODO so far we only create MDCT atoms with sinus windows
    Waveforms = {}
        # Waveforms are stored in a dictionary, the key is Length+Frequency

    Windows = {}  # Sine Windows : key is the scale

    # constructor is the same
    def __init__(self, useC=True):
        """ The Server is global"""
        self.use_c_optim = useC

    # no twiddling coefficient
    def initialize(self, L):
        """ instantiate the coefficients for the given scale """
        self.Windows[L] = math.sqrt(8.0 / (3.0 * L)) * hann(L)

    def get_waveform(self, scale, binIndex, phase):

        return parallelProjections.get_real_gabor_atom(int(scale), binIndex, phase)

#        # Check whether the waveform is already in the dictionary
#        key = scale + binIndex + float(phase)/(2*math.pi)
#        if not key in self.Waveforms:
#            self.Waveforms[key] = self.create_waveform(scale , binIndex , phase)
#
#        return self.Waveforms[key]

    def create_waveform(self, scale, binIndex, phase):
        if not scale in self.Windows:
            self.initialize(scale)
        wf = math.sqrt(2.0) * self.Windows[scale] * np.array([np.cos(
            binIndex * l + phase) for l in range(scale)])
        # scrap zeroes on the borders
        return wf



#*****************************************************************************/
#                                                                            */
#                                log.py                                      */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */
# */

'''
Module log
==========

'''

import logging
import traceback

Levels = {-1: logging.NOTSET,
          0: logging.ERROR,
          1: logging.WARNING,
          2: logging.DEBUG,
          3: logging.INFO}


class Log():
    """ This object should be made static and used to Log warnings and error messages
    debug Levels are handled that way:
        - level -1 : no messages at all
        - level 0 : only error messages are shown
        - level 1 : error and warning messages
        - level 2 : error , warnings and debug messages
        - level 3 : error , warnings , debug and info messages
    the imode attribute sets matplotlib drawing mode to interactive"""

    # attributes
    name = None
    outputFile = None
    debugLevel = 0
    imode = True
    logger = None
    noCtx = False

    # constructor
    def __init__(self, loggerName,  level=0, outputFilePath=None, imode=False, noContext=False):
        """ constructor """
        self.name = loggerName
        self.outputFile = outputFilePath
        self.debugLevel = level
        self.imode = imode
        self.noCtx = noContext
        # create logger
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(Levels[self.debugLevel])

        # create console handler and set level to debug
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # create formatter

        formatter = logging.Formatter(" %(levelname)s - %(message)s")

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.logger.addHandler(ch)

        # "application" code
        if self.outputFile is None:
            self.logger.info("created log handler with no file output for function : " +
                 self.name)
        else:
            self.logger.info("created log handler for function : " + self.name)

        if self.imode:
            import matplotlib.pyplot as plt
            plt.ion()
            self.logger.info("Set interactive mode ON for pyplot")

    def set_level(self, newLevel):
        self.debugLevel = newLevel
        self.logger.setLevel(Levels[self.debugLevel])

    def get_context(self):
        if self.noCtx:
            return ''
        ctx_list = traceback.extract_stack(limit=3)
        str_context = ctx_list[0][0].split('/')[-1] + '::' + str(ctx_list[0][
            1]) + '::' + ctx_list[0][2] + '::'
        return str_context

    def info(self, message):
        self.logger.info(self.get_context() + message)

    def debug(self, message):
        self.logger.debug(self.get_context() + message)

    def warning(self, message):
        self.logger.warning(self.get_context() + message)

    def error(self, message):
        self.logger.error(self.get_context() + message)



#*****************************************************************************/
#                                                                            */
#                               Command line for Pymp.py                     */
#                                                                            */
#                        Matching Pursuit Library                            */
#                                                                            */
# M. Moussallam                                             Mon Aug 16 2010  */
# -------------------------------------------------------------------------- */
# */

'''
Module mp_cmd
=============

PyPursuit command line arguments looks as follow:
             -h or --help : print this help

             -f [filePath] : specifies the path to the audiofile, only wav files supported for now

             -w : flag for writing the output (approx and recomposedSignal) to output files

             -o [outPutFilePath] : path for the output

             -m [maxAtomNumber] :  specifies the maximum number of selected atoms

             -s [SRR] : specifies the target Signal-To-Residual Ratio

             -d [dictionarySizes] : specifies the sizes: e.g -d 128,1024,8192

             -t or --type ['mp' | 'LOMP'] : specify the chosen decomposition algorithm

             --debug=[0|1|2|3] : debug level. Default is 0

             -a or --pad pad signal with zeroes , default is 1

             -p or --plot plots the approximation using matplotib

             -l [segmentDuration] : length of the segments

             Example::

                 >>> python MPcmd.py -f sndFile.wav -m 100 -s 10 -d 128,1024,8192 --debug=1 -a -p

'''
import matplotlib.pyplot as plt
import sys
import getopt

from PyMP import signals, approx
from PyMP.mdct import Dico, LODico
from PyMP.mdct.rand import SequenceDico
from PyMP import  mp



def usage():
    print """Thanks for using Py_Pursuit. All questions please : manuel.moussallam@telecom-paristech.fr
PyPursuit command line arguments looks as follow:
             -h or --help : print this help

             -f [filePath] : specifies the path to the audiofile, only wav files supported for now

             -w : flag for writing the output (approx and recomposedSignal) to output files

             -o [outPutFilePath] : path for the output

             -m [maxAtomNumber] :  specifies the maximum number of selected atoms

             -s [SRR] : specifies the target Signal-To-Residual Ratio

             -d [dictionarySizes] : specifies the sizes: e.g -d 128,1024,8192

             -t or --type ['mp' | 'LOMP'] : specify the chosen decomposition algorithm

             --debug=[0|1|2|3] : debug level. Default is 0

             -a or --pad pad signal with zeroes , default is 1

             -p or --plot plots the approximation using matplotib

             -l [segmentDuration] : length of the segments

             example : python mp_cmd.py -f sndFile.wav -m 100 -s 10 -d 128,1024,8192 --debug=1 -a -p

             """


def main(argv):

    try:
        opts, args = getopt.getopt(argv, "hk:apt:bf:wo:m:s:d:l:", [
            "help", "debug=", "pad", "plot", "type=", "pipe"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    # initialize default values
    global _debug
    _debug = 0
    file_path = ''
    write = False
    plot = False
    out_path = 'output.xml'
    max_it_num = 100
    target_srr = 10
    pad = True
    dec_algo = 'mp'
    scales = [2 ** j for j in range(7, 15)]
    matpipe = False
    seg_length = 10
    # argument parsing

    try:
        for opt, arg in opts:
    #        print opt, arg
            if opt in ("-h", "--help"):
                usage()
                sys.exit()

            elif opt in ("-b", "--pipe"):
                matpipe = True

            # set debug level
            elif opt == "--debug":
                _debug = int(arg)

            elif opt == '-f':
                file_path = arg

            elif opt == '-w':
                write = True

            elif opt in ("-p", "--plot"):
                plot = True

            elif opt == '-o':
                out_path = arg

            elif opt == '-m':
                max_it_num = int(arg)

            elif opt == '-s':
                target_srr = int(arg)

            elif opt == '-l':
                seg_length = float(arg)

            # tricky dictionary sizes reading
            elif opt == '-d':
                scales = []
#                print arg
                for size in arg.split(','):
#                    print size
                    scales.append(int(size))

            elif opt in ('-a', '--pad'):
                pad = True

            elif opt in ('-t , --type'):
                dec_algo = arg
    except:
        print "sorry wrong syntax "
        usage()
        sys.exit()

    # end of argument parsing : print some info if debug level is above 0
    if _debug > 0:
        print "Starting ", dec_algo, " on ", file_path, " target_srr: ", target_srr, " max_it_num: ", max_it_num, " on ", len(scales), "xMDCT dico"
        print "debug=", _debug, "- Write ", prtBool(
            write), " - Plot : ", prtBool(plot)

    ####### Load audio signal - TODO more options ##########"""
# originalSignal = py_pursuit_Signal.InitFromFile(file_path, forceMono=True,
# doNormalize=True)
    longSignal = signals.LongSignal(
        file_path, frameDuration=seg_length)

    # now crop the signal to an adequate length
# originalSignal.crop(0 , math.floor(originalSignal.length/max(scales)) *
# max(scales))
#
#    if _debug > 0:
#        print "Cropping signal to ", originalSignal.length

    ####### Build dictionary - TODO more options and more dictionaries ########
    if dec_algo == 'mp':
        dictionary = Dico(scales)
    elif dec_algo == 'LOMP':
        dictionary = LODico(scales)
    elif dec_algo == 'RSSMP':
        dictionary = SequenceDico(scales)
    else:
        print "unrecognized dictionary type for now, availables are mp and LOMP"
        sys.exit()

    # main Decomposition algorithm
    if _debug > 0:
        print target_srr, max_it_num

    # additional parameter: segment length analysis and overlapping
    #segmentOverlap = 0.25 # in per one
# segPad = math.floor(seg_length * originalSignal.samplingFrequency
# /max(scales)) * max(scales); # increment in samples
#
#    # calcul of the number of segment needed
#    n_seg = int(math.floor(originalSignal.length / segPad))
    segmentNumber = longSignal.n_seg

    if _debug > 0:
        print segmentNumber, " segments"

    # the max_it_num is fixed in all segments

    # initialize approx
# approx = py_pursuit_Approx.py_pursuit_Approx(dictionary, [], None,
# originalSignal.length + dictionary.getN(), originalSignal.samplingFrequency)

    approxs = mp.mp_long(longSignal, dictionary, target_srr, max_it_num,
                               debug=_debug - 1, pad=pad, output_dir='', write=False)[0]

#    for segidx in range(n_seg):
#        if _debug >0:
#            print "Starting work on segment " , segidx;
#
#        subSignal = longSignal.getSubSignal(segidx, 1,)
#
##        if segidx == 8:
##            _debug =2;
##        else:
##            _debug=1;
#        subApprox = mp.mp(subSignal,dictionary,target_srr, max_it_num,
# debug=_debug-1,pad=pad, silentFail=True)[0]
#        # add all atoms to overall approx
#        for atom in subApprox.atoms:
#            atom.timePosition += segPad*segidx
#
#            # BUGFIX on the approx length due to padding
##            approx.addAtom(atom)
# approx.atoms.append(atom) # no need to add it to the recomposed signal if no
# synthesis

    # Fusion the sub approximants
    approx = approx.fusion_approxs(approxs, unPad=pad)

    if write:
        if _debug > 0:
            print "Writing to output file :", out_path
        approx.original_signal = longSignal
        approx.dumpToDisk(out_path)
#        approx.writeToXml(out_path)

    if plot:
        plt.figure()
        approx.plot_tf()
        plt.show()

    # end of program
    if _debug > 0:
        print "Exiting"

    # addon for matlab interface: outputs all the coeff
    if matpipe:
        for atom in approx.atoms:
            print atom.length, atom.freq_bin, atom.time_position
    #sys.exit()


def prtBool(bool):
    if bool:
        return 'yes'
    else:
        return 'no'

if __name__ == "__main__":
    main(sys.argv[1:])




"""
Module PyMP.mdct.rand.dico
==========================

This file handle dctionaries that are used in Randomized Matching Pursuits
see [1] for details.

[1] M. Moussallam, L. Daudet, et G. Richard,
"Matching Pursuits with Random Sequential Subdictionaries"
Signal Processing, vol. 92, pp. 2532-2544 2012.

"""

import math
import numpy as np

from ..dico import Dico
from ...baserand import AbstractSequenceDico
from ... import log
from . import block


#global _Logger
_Logger = log.Log('SSMPDicos', level=0)


class SequenceDico( AbstractSequenceDico, Dico):
    """ This dictionary implements a sequence of subdictionaries that are shifted in time at each iteration in a pre-defined manner
        the shifts are controlled by the different blocks.

        Attributes:
        -----------        
        `sequence_type`: The type of time-shift sequence, 
                available choices are *scale*,*random*,*gaussian*,*binom*,*dicho*,*jump*,*binary* 
                default is *random* which use a uniform pseudo-random generator

        `nb_consec_sim`: Number of consecutive iterations with the same time-shift (default is 1)

    """

    # properties
    sequence_type = 'none'  # type of sequence , Scale , Random or Dicho
    it_num = 0  # memorizes the position in the sequence
    nb_consec_sim = 1
    # number of consecutive similar position
    nature = 'RandomSequentialMDCT'

    # constructor
    def __init__(self, sizes=[], seq_type='random', nbSame=1,
                 windowType=None,seed=None):
        
        super(SequenceDico, self).__init__(seq_type=seq_type, nbSame=nbSame, seed=seed)
                
        self.sizes = sizes        
        self.windowType = windowType
        
        
    def initialize(self, residual_signal):
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1

        for mdctSize in self.sizes:
            # check whether this block should optimize time localization or not
            self.blocks.append(block.SequenceBlock(mdctSize, residual_signal, randomType=self.
                sequence_type, nbSim=self.nb_consec_sim, windowType=self.windowType, seed=self.seed))

    def compute_touched_zone(self, previousBestAtom):
        # if the current time shift is about to change: need to recompute all
        # the scores
        if (self.nb_consec_sim > 0):
            if ((self.it_num + 1) % self.nb_consec_sim == 0):
                self.starting_touched_index = 0
                self.ending_touched_index = -1
            else:                
                self.starting_touched_index = 0
                self.ending_touched_index = -1
#                self.starting_touched_index = previousBestAtom.time_position - 2* previousBestAtom.length
#                self.ending_touched_index = self.starting_touched_index + 2.5 * previousBestAtom.length
        else:
            self.starting_touched_index = previousBestAtom.time_position - previousBestAtom.length / 2
            self.ending_touched_index = self.starting_touched_index + 1.5 * previousBestAtom.length

    def update(self, residualSignal, iteration_number=0, debug=0):
        self.max_block_score = 0
        self.best_current_block = None
        self.it_num = iteration_number
        # BUGFIX STABILITY
#        self.endingTouchedIndex = -1
#        self.startingTouchedIndex = 0

        for block in self.blocks:
            startingTouchedFrame = int(
                math.floor(self.starting_touched_index / (block.scale / 2)))
            if self.ending_touched_index > 0:
                endingTouchedFrame = int(math.floor(self.
                    ending_touched_index / (block.scale / 2))) + 1
                # TODO check this
            else:
                endingTouchedFrame = -1

            block.update(residualSignal,
                 startingTouchedFrame, endingTouchedFrame, iteration_number)

            if np.abs(block.max_value) > self.max_block_score:
#                self.maxBlockScore = block.getMaximum()
                self.max_block_score = np.abs(block.max_value)
                self.best_current_block = block



class StochasticDico(Dico):
    """ This dictionary implements Stochastic MP as described
        in Elad et al 2009 (see also Ferrando et al 2000)
        
        All products are computed but the best atom is chosen
        probabilistically. Many parallel pursuits are then averaged."""
    
    # properties
    nature = 'RandomStochasticMDCT'

    # constructor
    def __init__(self, sizes=[], seq_type='random', nbSame=1,
                 windowType=None,seed=None, sigma=1.0, debug_level=0):
        self.sequence_type = seq_type
        self.sizes = sizes
        self.nb_consec_sim = nbSame
        self.windowType = windowType
        self.seed = seed
        self.sigma = sigma
        self.debug_level = debug_level
        
        
    def initialize(self, residual_signal):
        self.blocks = []
        self.best_current_block = None
        self.starting_touched_index = 0
        self.ending_touched_index = -1

        for mdctSize in self.sizes:
            # check whether this block should optimize time localization or not
            self.blocks.append(block.StochasticBlock(mdctSize, residual_signal,
                                                     windowType=self.windowType,
                                                     seed=self.seed,
                                                     sigma = self.sigma,
                                                     debug_level=self.debug_level))

    # DO I need to restate it?
    def update(self, residualSignal, iteration_number=0, debug=0):
        self.max_block_score = 0
        self.best_current_block = None
        self.it_num = iteration_number
        # BUGFIX STABILITY
#        self.endingTouchedIndex = -1
#        self.startingTouchedIndex = 0

        for block in self.blocks:
            startingTouchedFrame = int(
                math.floor(self.starting_touched_index / (block.scale / 2)))
            if self.ending_touched_index > 0:
                endingTouchedFrame = int(math.floor(self.
                    ending_touched_index / (block.scale / 2))) + 1
                # TODO check this
            else:
                endingTouchedFrame = -1

            block.update(residualSignal,
                 startingTouchedFrame, endingTouchedFrame)

            if np.abs(block.max_value) > self.max_block_score:
#                self.maxBlockScore = block.getMaximum()
                self.max_block_score = np.abs(block.max_value)
                self.best_current_block = block



from os.path import join
import numpy
import os
import numpy.core

NUMPYDIR = os.path.dirname(numpy.core.__file__)
libraries = ['fftw3']
include_dirs = [os.path.join(NUMPYDIR, r'include/numpy'),'/usr/local/include',r'C:\Libs']
library_dirs = ['/usr/lib/openmpi/lib/openmpi/', "/usr/local/lib",r'C:\Libs']

import platform
print platform.architecture()

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    import platform
    print platform.architecture()[0]
    if (platform.architecture()[0] == '32bit'):
        plt_str = '-DX=1'
    else:
        plt_str = '-DX=2'

    config = Configuration('PyMP', parent_package, top_path)

    config.add_extension('parallelProjections',
                         libraries=libraries,
                         include_dirs=include_dirs,
                         library_dirs=library_dirs,
                         sources=[join('src', 'parProj.c'),
                                  join('src', 'parallelProjections.c')],
                         extra_compile_args=['-fopenmp', '-fPIC', '-DDEBUG=0',plt_str],
                         extra_link_args=['-lgomp',])

    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



"""

"""
import numpy as np
from PyMP.mdct import Dico, atom
from PyMP import Signal, approx

sig = Signal('../data/glocs.wav', debug_level=3)
print sig
print sig.data

# sig.plot()
# sig.write('newDestFile.wav')
# editing

print 'Before cropping Length of ', sig.length
sig.crop(0, 2048)
print 'After cropping Length of ', sig.length

sub_sig = sig[0:2048]
print sub_sig

new_sig = Signal(np.ones((8,)), 1)
new_sig.data
print "Padding"
new_sig.pad(4)
new_sig.data
print "De-Padding"
new_sig.depad(4)
new_sig.data


dico = Dico([128, 1024, 8192])
sig = Signal('../data/glocs.wav', mono=True)
app = approx.Approx(dico, [], sig)


app.add(atom.Atom(256, 1, 256, 10, 8000, 1))

app.compute_srr()
print app




#
#
#                       Classes.mdct.pymp_MDCTAtom
#
#
#
# M. Moussallam                             Created on Nov 12, 2012
# -----------------------------------------------------------------------
#
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA  02111-1307, USA.
#

"""
Module mdct.atom
====================

This class inherits from :class:`.Atom` and is used to represent and manipulate MDCT atoms.
:class:`Atom` Objects can either be constructed or recovered from Xml using :func:`fromXml`

"""

import numpy as np

from ..base import BaseAtom
from .. import win_server
from ..tools.mdct import imdct

try:
    from xml.dom.minidom import Document, Element
except ImportError:
    print "Atoms can be saved and recovered in xml format, but you XML library seems not available"

#global _PyServer
#_Logger
# Initializing the waveform server as a global variable
_PyServer = win_server.get_server()


class Atom(BaseAtom):
    """ MDCT atom class : implement real domain MDCT atoms of scale defined by the atom length

    An MDCT atom is defined for a length L, a frequency localization k and a frame parameter p by:

    .. math::

        \phi_{L,p,k}[n]=w_{L}[u]\sqrt{\\frac{2}{L}} \cos [ \\frac{\pi}{L} \left(u+ \\frac{L+1}{2}\\right) (k+\\frac{1}{2}) ]

    Additionnal attributes are:

            - `freq_bin` : MDCT frequency bin index (default is 0)

            - `frame` : the frame index of the atom

            - `reduced_frequency`: corresponding sampling frequency (default is 0)

            - `mdct_value` : the atom mdct coefficient

    Additionnal parameters for time-shifted atoms:

            - `time_shift`: time shift in samples related to the closest MDCT grid index*

            - `proj_score`: useful when atom is reprojected using say.. LOMP algorithm

    """

    # MDCT attibutes
    nature = 'MDCT'
    freq_bin = 0
    frame = 0
    reduced_frequency = 0
    mdct_value = 0.0
    time_position = 0;

    # for time-shift invariant atoms
    time_shift = None
    proj_score = None

    # constructor
    def __init__(self, scale=0, amp=0, timePos=0, freqBin=0, Fs=0, mdctCoeff=0):
        ''' Basic constructor setting atom parameters '''
        self.length = scale
        self.amplitude = amp
        self.time_position = timePos
        self.freq_bin = freqBin
        self.fs = Fs
        if self.length != 0:
            self.reduced_frequency = (
                float(self.freq_bin + 0.5) / float(self.length))
        self.mdct_value = mdctCoeff

    # Synthesis routine - always prefer the PyWinServer unless you want a
    # specific window to ba applied
    def synthesize(self, value=None):
        """ synthesizes the waveform
            Specifies the amplitude, otherwise it will be initialiazed as unit-normed """

#        global _PyServer
        binIndex = np.math.floor(self.reduced_frequency * self.length)
        if value is None:
            self.waveform = self.mdct_value * _PyServer.get_waveform(
                self.length, binIndex)
        else:
            self.waveform = value * _PyServer.get_waveform(
                self.length, binIndex)

    def synthesize_ifft(self, newValue=None):
        ''' DEPRECATED synthesis using Python fftw3 wrapper but no waveform server... a lot slower '''
        mdctVec = np.zeros(3 * self.length)
        if newValue is None:
            mdctVec[self.length + self.freq_bin] = self.mdct_value
        else:
            mdctVec[self.length + self.freq_bin] = newValue
        self.waveform = imdct(
            mdctVec, self.length)[0.75 * self.length: 1.75 * self.length]
        return self.waveform

    def __eq__(self, other):
        ''' overloaded equality operator, allows testing equality between atom objects '''
        if not isinstance(other, BaseAtom):
            return False
        return ((self.length == other.length) & (self.time_position == other.time_position) & (self.freq_bin == other.freq_bin))

    def __repr__(self):
        if self.time_shift is None:
            ts_str = 'None'
        else:
            ts_str = str(self.time_shift)
        return '''
MDCT Atom : length = %d value = %2.2f
frame = %d  time Position = %d   time shift = %s
frequency Bin = %d   Frequency = %2.2f Hz''' % (self.length,
                            self.mdct_value,
                             self.frame,
                             self.time_position,
                             ts_str,
                             self.freq_bin,
                             self.reduced_frequency * self.fs,)


    def inner_prod(self, otherAtom):
        """ DEPRECATED returns the inner product between current atom and the other one
            This method should be as fast as possible"""
        waveform1 = self.get_waveform()
        waveform2 = otherAtom.get_waveform()
        startIdx = max(
            self.frame * self.length, otherAtom.frame * otherAtom.length)
        stopIdx = min((self.frame + 1) * self.length, (otherAtom.
                                                       frame + 1) * otherAtom.length)

        if startIdx >= stopIdx:
            return 0
            # disjoint support

        start1 = startIdx - self.frame * self.length
        start2 = startIdx - otherAtom.frame * otherAtom.length
        duration = stopIdx - startIdx
        norm1 = np.sqrt(sum(waveform1 ** 2))
        norm2 = np.sqrt(sum(waveform2 ** 2))

        return (self.mdct_value * otherAtom.mdct_value) * sum(waveform1[start1: start1 + duration] * waveform2[start2: start2 + duration]) / (norm1 * norm2)
#        return sum(self.get_waveform() * otherAtom.get_waveform())

    def get_waveform(self):
        ''' Retrieve the atom waveform '''
        if self.waveform is not None:
            return self.waveform
        else:
            return self.synthesize_ifft()

    def copy(self):
        '''copycat routine '''
        copyAtom = Atom(self.length,
                        self.amplitude,
                        self.time_position,
                        self.freq_bin,
                        self.fs,
                        self.mdct_value)
        copyAtom.frame = self.frame
        copyAtom.proj_score = self.proj_score
        copyAtom.time_shift = self.time_shift
        if self.waveform is not None:
            copyAtom.waveform = self.waveform
        else:
            copyAtom.synthesize()
        return copyAtom

    def get_value(self):
        return self.mdct_value

    def get_tf_info(self, Fs):
        """ returns energy centroids and spread in the time frequency plane """
        L = self.length
        K = L / 2
        freq = self.freq_bin
        f = float(freq) * float(Fs) / float(L)
        bw = (float(Fs) / float(K))  # / 2
        p = float(self.time_position + K) / float(Fs)
        l = float(L - K / 2) / float(Fs) 
        return f, bw, p, l
#def fromXml(xmlNode):
#    ''' Construct an Object from the corresponding XML Node '''
#    if not isinstance(xmlNode, Element):
#        raise TypeError('Xml element not provided')
#
#    atom = Atom(int(xmlNode.getAttribute('length')),
#                1,
#                int(xmlNode.getAttribute('tP')),
#                int(xmlNode.getAttribute('fB')),
#                int(xmlNode.getAttribute('Fs')),
#                float(xmlNode.getAttribute('value')))
#
#    atom.frame = int(xmlNode.getAttribute('frame'))
#    if not xmlNode.getAttribute('time_shift') in ('None', ''):
#        atom.time_shift = int(xmlNode.getAttribute('time_shift'))
#    if not xmlNode.getAttribute('score') in ('None', ''):
#        atom.proj_score = float(xmlNode.getAttribute('score'))
#    return atom




import math
import numpy as np

from ... import log
from ...mdct import dico as mdct_dico
from ..rand import dico as random_dico
from . import block as joint_block

_Logger = log.Log('JointDicos', level=0)


class SetDico(mdct_dico.Dico):
    """ This class handles multiple dictionaries the best atom is selected
        when best explaining all the signals
        Then a refinement phase is performed late to determine the local
        time shifts and amplitudes
        This class handles a set of dictionaries, one for each Jointly
        decomposed signals update method just call update in all
        dictionaries, just like initialize, get_best_atom starts with
        retrieving all the projections, sums them all and then decide what
        atom is the best choice. Then it is locally adapted to each
        signal, so a list of atoms is returned instead of one
        """

    # parameters
    dictionaryList = []
    sizes = []
    tolerances = []
#    blocksList = []
    blocks = []
    best_current_block = None
    max_block_score_list = []
    nature = None  # this parameter controls the atom selection function: sum sums scores across signals
                    # prod : multiply them
                    # maximin : selects

    starting_touched_index = []
    ending_touched_index = []

    def __init__(self, sizes, useC=True, selectNature='sum', tol=None, nonLinear=False, params=None):
        " Create an set of dictionaries, with blocks and everything."
        self.use_c_optim = useC
        self.sizes = sizes
        self.type = type
        self.projections = None
        self.nature = selectNature
        self.isNL = nonLinear
        if tol is not None:
            self.tolerances = tol
        else:
            self.tolerances = [2] * len(self.sizes)

#        print self.sizes, self.tolerances

        if selectNature == 'median' or selectNature == 'penalized'or selectNature == 'weighted':
            self.isNL = True
            print "NL dico detected"
            self.params = params

    def initialize(self, residualSignalList):
        self.blocks = []
        self.best_current_block = None

        ######### Optimized version : 1 block to handle all projections with sam
        for mdctSize, tolerance in zip(self.sizes, self.tolerances):
#            print mdctSize , " Tolerance : " , tolerance
            if tolerance > 1:
#                print "Adaptation assumed"
                if not self.isNL:
                    self.blocks.append(joint_block.SetLOBlock(mdctSize,
                                                              residualSignalList,
                                                              useC=self.use_c_optim,
                                                              nature=self.nature,
                                                              tolerance=tolerance))
                else:
                    self.blocks.append(joint_block.SetNLLOBlock(mdctSize,
                                                                residualSignalList,
                                                                useC=self.use_c_optim,
                                                                nature=self.nature,
                                                                tolerance=tolerance,
                                                                lambd=self.params))
            else:
# print "Tolerance too weak to allow atom optimization: No adaptation
# assumed"
                self.blocks.append(joint_block.SetBlock(mdctSize,
                                                        residualSignalList,
                                                        useC=self.use_c_optim,
                                                        nature=self.nature,
                                                        tolerance=tolerance))

        self.starting_touched_index = [0] * len(residualSignalList)
        self.ending_touched_index = [-1] * len(residualSignalList)

    def update(self, residualSignalList, iteratioNumber=0, debug=0):
        # Update all the blocks : Size by Size: so we can add the projections
        # and have find the maximum
        self.max_block_score = 0
        self.best_current_block = None

        startingTouchedFrameList = [0] * len(residualSignalList)
        endingTouchedFrameList = [-1] * len(residualSignalList)

        # if parrallel library not available
        for block in self.blocks:

            for sigIdx in range(len(residualSignalList)):

#                startingTouchedFrameList[sigIdx] = 0
#                endingTouchedFrameList[sigIdx] = -1
#                print "DEBUG : recomputing all"
                startingTouchedFrameList[sigIdx] = int(math.floor(
                    self.starting_touched_index[sigIdx] / (block.scale / 2)))

                if self.ending_touched_index[sigIdx] > 0:
                    endingTouchedFrameList[sigIdx] = int(math.floor(self.ending_touched_index[sigIdx] / (block.scale / 2))) + 1  # TODO check this
                else:
                    endingTouchedFrameList[sigIdx] = -1

# print "block: " , block.scale , " : " , startingTouchedFrame ,
# endingTouchedFrame
            block.update(residualSignalList,
                         startingTouchedFrameList, endingTouchedFrameList)

            if abs(block.max_value) > self.max_block_score:
                self.max_block_score = abs(block.max_value)
                self.best_current_block = block

    def get_best_atom(self, debug, noAdapt=False):
        if self.best_current_block is None:
            score = 0
            for block in self.blocks:
                maxValue = max(block.best_score_tree)
                if abs(maxValue) > score:
                    self.max_block_score = abs(maxValue)
                    self.best_current_block = block

        if self.best_current_block is None:
            raise ValueError("no best block constructed, make sure inner product have been updated")

# print 'Best Size is ', self.sizes[self.bestSize] , ' with score ' ,
# self.bestScore , ' for atom ', self.maxAtomIdx

        # call on the best block to return a list of atoms that are adapted to
        # each of the signals
        self.bestAtoms = self.best_current_block.get_optimized_best_atoms(
            debug, noAdapt=False)
#        for sigIdx in range(len(self.blocksList)):
# self.bestAtoms.append(self.blocksList[sigIdx][self.bestSize].getAdaptedMaxAtom(self.maxAtomIdx))

        return self.bestAtoms

    def get_mean_atom(self, getFirstAtom=True):
        ''' retrieve a mean best atom: with mean position and amplitude '''

        if self.bestAtoms is None or len(self.bestAtoms) < 1:
            print " Empty set of Atoms: cannot create template"

        self.meanAtom = self.bestAtoms[0].copy()

        if getFirstAtom:
            return self.meanAtom

        value = 0.0
        timePos = 0

#        for atom in self.bestAtoms:
##            value += sqrt(abs(atom.getAmplitude()))
#            timePos += atom.timePosition
        value = np.mean([abs(at.proj_score) for at in self.bestAtoms])
        timePos = np.median([at.time_position for at in self.bestAtoms])

#        self.meanAtom.timePosition = int(timePos)#/len(self.bestAtoms)
        self.meanAtom.mdct_value = float(value)

        self.meanAtom.waveform /= np.sqrt(np.sum(self.meanAtom.waveform ** 2))
        self.meanAtom.waveform *= float(value)  # /float(len(self.bestAtoms))

#        print "Mean Value of ",float(value)#/float(len(self.bestAtoms))
        return self.meanAtom

    def compute_touched_zone(self, sigIdx, atom):
        # print "Updating : " ,blockIdx
        # Each block need be recomputed on a different fraction of the signal
        if atom is not None:
            self.starting_touched_index[
                sigIdx] = atom.time_position - atom.length / 2
            self.ending_touched_index[
                sigIdx] = atom.time_position + 1.5 * atom.length
        # if the atom is not selected , then no update is necessary
        else:
            self.ending_touched_index[
                sigIdx] = self.starting_touched_index[sigIdx]


class RandomSetDico(SetDico, random_dico.SequenceDico):
    """ This class handles multiple dictionaries the best atom is selected when best explaining all the signals
        Then a refinement phase is performed late to determine the local time shifts and amplitudes
        This class handles a set of dictionaries, one for each Jointly decomposed signals
        update method just call update in all dictionaries, just like initialize
        get_best_atom starts with retrieving all the projections, sums them all and then decide what atom is the
        best choice. Then it is locally adapted to each signal, so a list of atoms is returned instead of one

        USES RSSMP

        """

    # parameters
    dictionaryList = []
    sizes = []
    tolerances = []
#    blocksList = []
    blocks = []
    best_current_block = None
    max_block_score_list = []
    nature = None  # this parameter controls the atom selection function: sum sums scores across signals
                    # prod : multiply them
                    # maximin : selects

    starting_touched_index = []
    ending_touched_index = []
    TsSequence = None

    def __init__(self, sizes, useC=True, selectNature='sum',
                 tol=None, nonLinear=False, params=None, seed=None):
        " Create an set of dictionaries, with blocks and everything."
        self.use_c_optim = useC
        self.sizes = sizes
        self.type = type
        self.projections = None
        self.nature = selectNature
        self.isNL = nonLinear
        self.seed = seed
        if tol is not None:
            self.tolerances = tol
        else:
            self.tolerances = [2] * len(self.sizes)

#        print self.sizes, self.tolerances

        if selectNature == 'median' or selectNature == 'penalized'or selectNature == 'weighted':
            self.isNL = True
            print "NL dico detected"
            self.params = params

    def initialize(self, residualSignalList):
        self.blocks = []
        self.best_current_block = None

        ######### Optimized version : 1 block to handle all projections with sam
        for mdctSize, tolerance in zip(self.sizes, self.tolerances):
#            print mdctSize , " Tolerance : " , tolerance
#            if tolerance > 1:
#                if not self.isNL:
#                    self.blocks.append(Block.py_pursuit_SetCCBlock(mdctSize ,
#                                                               residualSignalList,
#                                                                useC=self.useC ,
#                                                                nature= self.nature,
#                                                                tolerance = tolerance ))
#                else:
#                    self.blocks.append(Block.py_pursuit_SetNLCCBlock(mdctSize ,
#                                                               residualSignalList,
#                                                                useC=self.useC ,
#                                                                nature= self.nature,
#                                                                tolerance = tolerance ,
#                                                                lambd = self.params))
#            else:
# print "Tolerance too weak to allow atom optimization: No adaptation
# assumed"
            self.blocks.append(joint_block.RandomSetBlock(mdctSize,
                                                          residualSignalList,
                                                          useC=self.use_c_optim,
                                                          nature=self.nature,
                                                          tolerance=tolerance,
                                                          seed=self.seed))

        self.starting_touched_index = [0] * len(residualSignalList)
        self.ending_touched_index = [-1] * len(residualSignalList)

    def update(self, residualSignalList, iteratioNumber=0, debug=0):
        # Update all the blocks : Size by Size: so we can add the projections
        # and have find the maximum
        self.max_block_score = 0
        self.best_current_block = None

        startingTouchedFrameList = [0] * len(residualSignalList)
        endingTouchedFrameList = [-1] * len(residualSignalList)

        # if parrallel library not available
        for block in self.blocks:

#            for sigIdx in range(len(residualSignalList)):
#
##                startingTouchedFrameList[sigIdx] = 0
##                endingTouchedFrameList[sigIdx] = -1
##                print "DEBUG : recomputing all"
#                startingTouchedFrameList[sigIdx] = int(math.floor(self.startingTouchedIndex[sigIdx] / (block.scale/2)))
#
#                if self.endingTouchedIndex[sigIdx] > 0:
#                    endingTouchedFrameList[sigIdx] = int(math.floor(self.endingTouchedIndex[sigIdx] / (block.scale/2) )) + 1 # TODO check this
#                else:
#                    endingTouchedFrameList[sigIdx] = -1

# print "block: " , block.scale , " : " , startingTouchedFrame ,
# endingTouchedFrame
            block.update(residualSignalList, startingTouchedFrameList,
                         endingTouchedFrameList, iteratioNumber)

            if abs(block.max_value) > self.max_block_score:
                self.max_block_score = abs(block.max_value)
                self.best_current_block = block




#
#                       base classes
#
#
#
# M. Moussallam                             Created on Dec 27, 2012
# -----------------------------------------------------------------------
#
#
import numpy as np

class BaseSignal(object):
    """
    Attributes
    ----------
    *data* : array
        a numpy array containing the signal data
    *fs* : int
        The sampling frequency
    *location* :  str , optional
        Where the original file is located on the disk
    """
    
    data = np.array([])
    fs = 0
    location = ""

    def __init__(self):
        """ default constructor doesn't do anything """
    
    def add(self, atom):
        raise NotImplementedError("Subclass method not found")
        
    def subtract(self, atom):       
        raise NotImplementedError("Subclass method not found") 


class BaseAtom(object):
    """ Abstract Atom class interface:

    To implement a new type of atom, you must derive from this class.

    Attributes
    ----------
    nature: str
        A string describing the atom type (e.g MDCT, MCLT , GaborReal) default is MDCT
    waveform : array-like
        a numpy array that will contain the atom waveform
    amplitude : float
        the atom amplitude
    fs : int 
        the atom sampling frequency

    """

    # default values
    nature = 'Abstract'    
    waveform = None
    length = 0
    fs = 0
    amplitude = 0    

    def __init__(self):
        """" basic constructor doesn't do anything"""

    # mandatory functions
    def get_waveform(self):
        # A function to retrieve the atom waveform
        return self.waveform


class BaseBlock(object):
    ''' A block is an instance handling projections for Matching Pursuit.

        Mandatory fields:
            - type : the type of dictionary (e.g  Gabor , MDCT , Haar ...)
            - scale : the scale of the block
            - residualSignal : a py_pursuit_Signal instance that describes the current residual

        Mandatory methods
            - update :  updates the inner products table
            - getMaximum : retrieve the maximum absolute value of inner products
            - getMaxAtom : return a corresponding Atom instance'''

    # members
    scale = 0
    residual_signal = BaseSignal()

    # methods
    def __init__(self):
        """ empty constructor """

    def update(self, residual):
        raise NotImplementedError("Subclass method not found")

    def find_max(self):
        raise NotImplementedError("Subclass method not found")

    def get_max_atom(self):
        raise NotImplementedError("Subclass method not found")

class BaseDico(object):
    """ This class creates an interface that any type of dictionary should reproduce
        in order to be used correclty by Pursuit algorithm in this framework:

            - `sizes`: a list of scales

            - `blocks`: a list of :class:`.BaseBlock` that handles the projection of a residual signal along with
                        the selection of a projection maximum given a transform and a criteria
        """
    # attributes:
    sizes = None    
    blocks = []    
    nature = 'Abstract'    

    def __init__(self):
        """ default constructor doesn't do anything"""

    def get_pad(self):
        """ the amount of zeroes that need to be added on the sides """
        return self.sizes[-1]  # last element of the list should be the biggest, 

    def update(self, res, it):
        raise NotImplementedError("Subclass method not found")

    def get_best_atom(self):
        raise NotImplementedError("Subclass method not found")
    
                        
class BaseApprox(object):
    """
    Attributes
    ----------
    *dico* : :class:`.BaseDico`
        the dictionary (as a :class:`.BaseDico` object) from which it has been constructed
    *atoms* : list
        a list of :class:`.BaseAtom` objets
    *original_signal* : :class:`.BaseSignal`
        a :class:`.Signal` object that is the original signal
    *recomposed_signal* :  :class:`.BaseSignal`
        a :class:`.BaseSignal` objet for the reconstructed signal (as the weighted sum of atoms specified in the atoms list)
    """
    dico = BaseDico()
    atoms = []
    original_signal = BaseSignal()
    recomposed_signal = BaseSignal()
    
    def __init__(self):
        """ default constructor doesn't do anything """
    
    def add(self, atom):
        raise NotImplementedError("Subclass method not found")
        
    def remove(self, atom):       
        raise NotImplementedError("Subclass method not found")   
    
 


"""

Tutorial provided as part of PyMP

M. Moussallam

"""
from PyMP.mdct import Dico, LODico
from PyMP.mdct.rand import SequenceDico
from PyMP import mp, mp_coder, Signal
signal = Signal('../data/ClocheB.wav', mono=True)  # Load Signal
signal.crop(0, 4.0 * signal.fs)     # Keep only 4 seconds
# atom of scales 8, 64 and 512 ms
scales = [(s * signal.fs / 1000) for s in (8, 64, 512)]
signal.pad(scales[-1])
# Dictionary for Standard MP
dico = Dico(scales)
# Launching decomposition, stops either at 20 dB of SRR or 2000 iterations
app, dec = mp.mp(signal, dico, 20, 2000, pad=False)

app.atom_number

snr, bitrate, quantized_app = mp_coder.simple_mdct_encoding(
    app, 8000, Q=14)
print (snr, bitrate)

print "With Q=5"
snr, bitrate, quantized_app = mp_coder.simple_mdct_encoding(
    app, 8000, Q=5)
print (snr, bitrate)


snr, bitrate, quantized_app = mp_coder.simple_mdct_encoding(
    app, 2000, Q=14)
print (snr, bitrate)

lomp_dico = LODico(scales)
lomp_app, lomp_dec = mp.mp(
    signal, lomp_dico, 20, 2000, pad=False)
lomp_snr, lomp_bitrate, lomp_quantized_app = mp_coder.simple_mdct_encoding(
    lomp_app, 2000, Q=14, shift_penalty=True)
print (lomp_snr, lomp_bitrate)

rssmp_dico = SequenceDico(scales)
rsssmp_app, rssmp_dec = mp.mp(signal, rssmp_dico, 20, 2000, pad=False)
rssmp_snr, rssmp_bitrate, rssmp_quantized_app = mp_coder.simple_mdct_encoding(
    rsssmp_app, 2000, Q=14)
print (rssmp_snr, rssmp_bitrate)

print (quantized_app.atom_number, lomp_quantized_app.atom_number,
       rssmp_quantized_app.atom_number)

# quantized_app.plotTF()
# plt.show()


print " now at a much larger level : a snr of nearly 40 dB and around 20 Kbps"
app, dec = mp.mp(signal, dico, 40, 20000, pad=False)
snr, bitrate, quantized_app = mp_coder.simple_mdct_encoding(
    app, 64000, Q=16)
print (snr, bitrate)

del app, dico

lomp_app, lomp_dec = mp.mp(
    signal, lomp_dico, 40, 20000, pad=False)
lomp_snr, lomp_bitrate, lomp_quantized_app = mp_coder.simple_mdct_encoding(
    lomp_app, 64000, Q=16, shift_penalty=True)
print (lomp_snr, lomp_bitrate)

del lomp_app, lomp_dico

rsssmp_app, rssmp_dec = mp.mp(
    signal, rssmp_dico, 40, 20000, pad=False)
rssmp_snr, rssmp_bitrate, rssmp_quantized_app = mp_coder.simple_mdct_encoding(
    rsssmp_app, 64000, Q=16)
print (rssmp_snr, rssmp_bitrate)
